// Copyright (C) 2009 Playedit.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "config.h"

#include "mtp_device.h"
#include "playlist.h"
#include "music.h"

#include <glibmm/main.h>
#include <gdkmm/devicemanager.h>
#include <iostream>

const unsigned int EXPAND_TIMEOUT = 1000;
const unsigned int SCROLL_TIMEOUT = 100;
const double SCROLL_EDGE_SIZE = 25;

// Expand the row on drag hover
bool PlaylistView::Expand()
{
	Gtk::TreePath path;
	Gtk::TreeViewDropPosition pos;
	get_drag_dest_row(path, pos);
	if (path) expand_row(path, false);

	return false;
}

// Automatically scroll up or down if a drag hovers near the beginning or end
// of the treeview
bool PlaylistView::AutoScroll()
{
	bool scroll_on = true;
	int x, y;
	Gdk::ModifierType mask;

	get_bin_window()->get_device_position(
			Gdk::Display::get_default()->get_device_manager()->get_client_pointer(), x, y, mask);

	Glib::RefPtr<Gtk::Adjustment> v_adj = get_vadjustment();
	int orig_adj_value = v_adj->get_value();

	if ((y > v_adj->get_page_size() - SCROLL_EDGE_SIZE) &&
			(v_adj->get_value() + v_adj->get_page_size() < v_adj->get_upper()))
		v_adj->set_value(std::min(v_adj->get_value() + (SCROLL_EDGE_SIZE + y - v_adj->get_page_size()),
					v_adj->get_upper() - v_adj->get_page_size()));
	else if ((y < SCROLL_EDGE_SIZE) && (v_adj->get_value() > 0))
		v_adj->set_value(std::max(v_adj->get_value() - (SCROLL_EDGE_SIZE - y), v_adj->get_lower()));

	Gtk::TreeView::on_drag_leave(Glib::RefPtr<Gdk::DragContext>(), GDK_CURRENT_TIME);

	// Don't scroll outside of a playlist that a drag comes from
	if (!m_bScrollOutOfPlaylist)
	{
		int page_height;
		convert_bin_window_to_widget_coords(0, v_adj->get_page_size(), x, page_height);
		y = (y < 24 ? 25 : (y < page_height ? y : page_height - 1));

		std::vector<Gtk::TreePath> selected_paths = get_selection()->get_selected_rows();
		Gtk::TreePath path;
		Gtk::TreeViewDropPosition pos;
		if (selected_paths.size() && get_dest_row_at_pos(1, y, path, pos))
			scroll_on = (path[0] == selected_paths.front()[0]);
	}

	return (orig_adj_value != v_adj->get_value()) && scroll_on;
}

// When a track is drug in highlight the row where the track will be dropped.
// Also expand a parent if hovered over with a drag.
bool PlaylistView::on_drag_motion(const Glib::RefPtr<Gdk::DragContext>& context,
		int x, int y, guint time)
{
	bool ret = false;

	Gtk::TreePath path;
	Gtk::TreeViewDropPosition pos = Gtk::TREE_VIEW_DROP_BEFORE;

	if (get_dest_row_at_pos(x, y, path, pos))
	{
		// Don't allow drags from one playlist to another
		Gtk::Widget* source_treeview = drag_get_source_widget(context);
		if (source_treeview == this)
		{
			m_bScrollOutOfPlaylist = false;
			std::vector<Gtk::TreePath> selected_paths =
				get_selection()->get_selected_rows();
			if (selected_paths.size())
				ret = (selected_paths.front()[0] == path[0]);
		}
		else
		{
			m_bScrollOutOfPlaylist = true;
			ret = true;
		}
		if (ret)
		{
			if (path.size() == 2)
			{
				// Set the highlight
				if ((pos == Gtk::TREE_VIEW_DROP_BEFORE)
						|| (pos == Gtk::TREE_VIEW_DROP_INTO_OR_BEFORE))
					pos = Gtk::TREE_VIEW_DROP_BEFORE;
				else
					pos = Gtk::TREE_VIEW_DROP_AFTER;
				set_drag_dest_row(path, pos);

				// Remove hover expand timeout if there is one
				if (expand_timeout.connected()) expand_timeout.disconnect();
			}
			else
			{
				Gtk::TreePath last_path;
				get_drag_dest_row(last_path, pos);
				// Remove hover expand timeout if the path has changed
				if ((!last_path || last_path != path) && expand_timeout.connected())
					expand_timeout.disconnect();

				// Add the hover expand timeout if there are children, the row
				// is not expanded, and the drag comes from the other tree
				if (source_treeview != this &&
						m_Store->get_iter(path)->children().size() &&
						!row_expanded(path) &&
						expand_timeout.empty())
					expand_timeout = Glib::signal_timeout().connect(
							sigc::mem_fun(*this, &PlaylistView::Expand),
							EXPAND_TIMEOUT);

				// Set the highlight on the playlist
				set_drag_dest_row(path, Gtk::TREE_VIEW_DROP_INTO_OR_AFTER);
			}

			// Set up auto scrolling
			int bx, by;
			convert_widget_to_bin_window_coords(0, y, bx, by);
			Glib::RefPtr<Gtk::Adjustment> v_adj = get_vadjustment();
			if (((by > v_adj->get_page_size() - SCROLL_EDGE_SIZE)
						&& (v_adj->get_value() + v_adj->get_page_size() < v_adj->get_upper()))
					|| ((by < SCROLL_EDGE_SIZE) && (v_adj->get_value() > 0)))
			{
				if (auto_scroll_timeout.empty())
					auto_scroll_timeout = Glib::signal_timeout().connect(
							sigc::mem_fun(*this, &PlaylistView::AutoScroll),
							SCROLL_TIMEOUT);
			}
			else
				if (auto_scroll_timeout.connected()) auto_scroll_timeout.disconnect();
		}
		else
		{
			if (auto_scroll_timeout.connected()) auto_scroll_timeout.disconnect();
			Gtk::TreeView::on_drag_leave(Glib::RefPtr<Gdk::DragContext>(), time);
		}
	}
	else
	{
		if (expand_timeout.connected()) expand_timeout.disconnect();
		Gtk::TreeView::on_drag_leave(Glib::RefPtr<Gdk::DragContext>(), time);
	}

	queue_draw();

	if (!ret) context->drag_refuse(time);
	return ret;
}

// Remove expand timeout if it still exists and clear the destination row when
// the mouse drag leaves.
void PlaylistView::on_drag_leave(const Glib::RefPtr<Gdk::DragContext>& context, guint time)
{
	// Remove time outs if there are any
	if (expand_timeout.connected()) expand_timeout.disconnect();
	Gtk::TreeView::on_drag_leave(context, time);
	queue_draw();
}

// Remove the autoscroll timeout if the drag ends while scrolling
void PlaylistView::on_drag_end(const Glib::RefPtr<Gdk::DragContext>& context)
{
	if (auto_scroll_timeout.connected()) auto_scroll_timeout.disconnect();
	Gtk::TreeView::on_drag_end(context);
}

// Tracks have been dropped on the playlist treeview.  Add tracks to the
// playlist dropped on if they came from the music treeview, or move track to
// new position if it came from the playlist treeview.
void PlaylistView::on_drag_data_received(const Glib::RefPtr<Gdk::DragContext>& context,
		int x, int y, const Gtk::SelectionData& /* selection_data */, guint /* info */, guint /* time */)
{
	Gtk::TreePath drop_path;
	Gtk::TreeViewDropPosition pos = Gtk::TREE_VIEW_DROP_BEFORE;

	if (get_dest_row_at_pos(x, y, drop_path, pos))
	{
		guint new_position;
		Gtk::TreeIter drop_iter = m_Store->get_iter(drop_path);
		Gtk::TreeIter drop_iter_parent = drop_iter->parent();
		if (drop_iter_parent)
		{
			if (pos == Gtk::TREE_VIEW_DROP_BEFORE ||
					pos == Gtk::TREE_VIEW_DROP_INTO_OR_BEFORE)
			{
				pos = Gtk::TREE_VIEW_DROP_BEFORE;
				new_position = drop_path[1];
			}
			else
			{
				pos = Gtk::TREE_VIEW_DROP_AFTER;
				new_position = drop_path[1] + 1;
			}
		}
		else
		{
			pos = Gtk::TREE_VIEW_DROP_INTO_OR_AFTER;
			drop_iter_parent = drop_iter;
			new_position = drop_iter->children().size();
		}

		Glib::RefPtr<playlist_t> playlist =
			Glib::RefPtr<playlist_t>::cast_static(drop_iter_parent->get_value(m_StoreColumns.m_Data));

		Gtk::TreeView* source_view = (Gtk::TreeView*)drag_get_source_widget(context);
		Glib::RefPtr<Gtk::TreeModel> source_model = source_view->get_model();
		if (source_model == m_Store)
		{
			// Move tracks in playlist
			// Get the currently selected paths in the playlist
			std::vector<Gtk::TreePath> source_paths = get_selection()->get_selected_rows();

			// Move the tracks to the new position
			std::vector<uint32_t> new_tracks = playlist->Tracks();
			std::vector<uint32_t> moved_tracks;
			for (std::vector<Gtk::TreePath>::iterator cur_path = source_paths.begin();
					cur_path != source_paths.end(); ++cur_path)
			{
				if ((guint)(*cur_path)[1] < new_position)
					--new_position;
				moved_tracks.push_back(new_tracks[(*cur_path)[1] - moved_tracks.size()]);
				new_tracks.erase(new_tracks.begin() + (*cur_path)[1] - moved_tracks.size() + 1);
			}
			new_tracks.insert(new_tracks.begin() + new_position, moved_tracks.begin(), moved_tracks.end());

			playlist->SetTracks(new_tracks);

			// Update playlist on device
			if (m_Device.UpdatePlaylist(playlist))
			{
				// Convert the source paths to row references
				std::vector<Gtk::TreeRowReference> refs = ConvertPathsToRowRefs(source_paths);

				// Move the tracks in the playlist store
				for (std::vector<Gtk::TreeRowReference>::iterator ref = refs.begin();
						ref != refs.end(); ++ref)
				{
					if (pos == Gtk::TREE_VIEW_DROP_AFTER)
						m_Store->move(m_Store->get_iter(ref->get_path()), ++drop_iter);
					else if (pos == Gtk::TREE_VIEW_DROP_BEFORE)
						m_Store->move(m_Store->get_iter(ref->get_path()), drop_iter);
					else
						m_Store->move(m_Store->get_iter(ref->get_path()), drop_iter_parent->children().end());
					drop_iter = m_Store->get_iter(ref->get_path());
					pos = Gtk::TREE_VIEW_DROP_AFTER;
				}
			}
			else
				// Update didn't work.  Reset the tracks
				playlist->SetTracks(m_Device.GetPlaylist(playlist->Id())->Tracks());
		}
		else
		{
			// Add tracks to playlist dropped into.
			std::vector< Glib::RefPtr<track_t> > new_tracks;

			// First get the tracks that are going to be added.
			std::vector<Gtk::TreePath> source_paths = source_view->get_selection()->get_selected_rows();
			if (source_paths.front().size() == 1)
			{
				for (std::vector<Gtk::TreePath>::iterator source_path = source_paths.begin();
						source_path != source_paths.end(); ++source_path)
				{
					Gtk::TreeNodeChildren album = source_model->get_iter(*source_path)->children();
					for (Gtk::TreeIter iter = album.begin();
							iter != album.end(); ++iter)
					{
						Glib::RefPtr<track_t> album_track = Glib::RefPtr<track_t>::cast_static(
								iter->get_value(m_StoreColumns.m_Data));
						new_tracks.push_back(album_track);
					}
				}
			}
			else if (source_paths.front().size() == 2)
			{
				for (std::vector<Gtk::TreePath>::iterator source_path = source_paths.begin();
						source_path != source_paths.end(); ++source_path)
				{
					Glib::RefPtr<track_t> tmp_track = Glib::RefPtr<track_t>::cast_static(
							source_model->get_iter(*source_path)->get_value(m_StoreColumns.m_Data));
					new_tracks.push_back(tmp_track);
				}
			}

			// Add the tracks to the playlist
			std::vector<uint32_t> tracks = playlist->Tracks();
			for (std::vector< Glib::RefPtr<track_t> >::reverse_iterator new_track = new_tracks.rbegin();
					new_track != new_tracks.rend(); ++new_track)
				tracks.insert(tracks.begin() + new_position, (*new_track)->Id());
			playlist->SetTracks(tracks);

			// Update playlist on device
			if (m_Device.UpdatePlaylist(playlist))
			{
				// Add the tracks to the playlist store
				for (std::vector< Glib::RefPtr<track_t> >::reverse_iterator track = new_tracks.rbegin();
						track != new_tracks.rend(); ++track)
				{
					Gtk::TreeIter new_iter = m_Store->insert(drop_iter_parent->children()[new_position]);
					(*new_iter)[m_StoreColumns.m_Data] = *track;
				}
				if ((pos == Gtk::TREE_VIEW_DROP_INTO_OR_AFTER) && !row_expanded(drop_path))
					expand_row(drop_path, false);
			}
			else
				// Update didn't work.  Reset the tracks
				playlist->SetTracks(m_Device.GetPlaylist(playlist->Id())->Tracks());
		}
	}
}
